using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;

namespace Feel
{
    public class GlassBody : MVRScript
    {
        #region PluginInfo    
        public string pluginAuthor = "Feel";
        public string pluginName = "GlassBody";
        public string pluginVersion = "1";
        public string pluginDate = "04/11" + "/2022";
        public string pluginDescription = @"Glass Body";
        #endregion

        #region Vars    

        protected bool first_loading = false;

        protected JSONStorableFloat PushMaxLevel;
        protected JSONStorableFloat PushLimLevel;
        protected JSONStorableBool IsMovie;
        protected JSONStorableFloat PushSpeed;
        protected JSONStorableFloat MovieFrameValue;
        protected JSONStorableFloat GenitaliaScaleValue;
        protected JSONStorableFloat GenitaliaPowerValue;
        protected JSONStorableFloat GenitaliaRocValue;
        protected JSONStorableBool BodyAngleLock;
        protected JSONStorableFloat BodyAlphaValue;
        protected JSONStorableColor BodyColorValue;
        protected JSONStorableBool IsHeadActive;
        protected JSONStorableFloat headXRotation;
        protected JSONStorableFloat headYRotation;
        protected JSONStorableFloat headZRotation;

        protected List<String> _atomBodyChoices;
        protected JSONStorableStringChooser _personSelectJSON;
        protected List<String> _targetBodyChoices;
        protected UIDynamicPopup _personDisplayPopup;
        protected Atom _selectAtom;
        protected JSONStorableFloat _bodyPosValue;
        protected JSONStorableFloat _bodyYCorrection;
        protected JSONStorableFloat BodyRotationValue;

        protected JSONStorableStringChooser _headTargetSelectJSON;
        protected List<String> _headTargetBodyChoices;
        protected UIDynamicPopup _headTargetDisplayPopup;

        private Rigidbody _lipTrigger;
        private Rigidbody _mouthTrigger;
        private Rigidbody _labiaTrigger;
        private Rigidbody _vaginaTrigger;

        private Atom _thisAtom;
        private FreeControllerV3 _thisControl;

        protected List<String> _orificePersonChoices;

        protected JSONStorableStringChooser _orificeJSON;
        protected UIDynamicPopup _displayPopup;

        public Transform ObjectRoot;

        public Transform ArmatureRoot;
        public Transform LegRoot;
        public Transform AbdomenRoot;
        public Transform ChestRoot;
        public Transform GenitaliaRoot;
        public Transform Gen1Root;
        public Transform Gen2Root;
        public Transform Gen3Root;
        public Transform Gen4Root;
        public Transform Gen5Root;
        public Transform Gen6Root;
        public Transform Gen7Root;
        public Transform Gen8Root;
        public Transform BodyMeshRoot;
        public Transform HeadTargetRoot;
        public Transform CumHeadRoot;

        private Vector3[] base_pos = new Vector3[8];

        private bool editorInitialized = false;
        private int initLoops = 0;
        private int maxInitLoops = 1500; // 3600 for one minute

        private bool is_movie_up = false;
        private float curr_movie_value = 0f;

        #endregion

        public override void Init()
        {
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("GlassBody- Please add this script on a Custom Unity Asset ( script disabled).");
                    return;
                }

                _thisAtom = containingAtom;
                _thisControl = _thisAtom.freeControllers.First(fc => fc.name == "control");

                _targetBodyChoices = new List<string>();

                _orificePersonChoices = new List<string>();
                _orificePersonChoices.Add("Mouth");
                _orificePersonChoices.Add("Vagina");
                _orificePersonChoices.Add("Anus");

                // A tiny bit of help to give some info to your fellow creator and/or player
                JSONStorableString movieTitleText = new JSONStorableString("MovieControl",
                    "<color=#000><size=35><b>Movie Control</b></size></color>");
                UIDynamic movieTitleTextfield = CreateTextField(movieTitleText, false);
                movieTitleTextfield.height = 30.0f;

                PushMaxLevel = new JSONStorableFloat("Push Max Level", 0f, OnSliderMaxChanged, 0f, 1.5f);
                RegisterFloat(PushMaxLevel);
                CreateSlider(PushMaxLevel, false);

                PushLimLevel = new JSONStorableFloat("Push Lim Level", 0f, OnSliderLimChanged, 0f, 0.5f);
                RegisterFloat(PushLimLevel);
                CreateSlider(PushLimLevel, false);

                IsMovie = new JSONStorableBool("Push Movie", false, ChangePushMovie);
                RegisterBool(IsMovie);
                UIDynamicToggle toggle2 = CreateToggle(IsMovie, false);
                toggle2.label = "Push Movie On Off";
                toggle2.height = 80;

                PushSpeed = new JSONStorableFloat("Push Speed", 1f, OnSliderSpeedChanged, 0.02f, 1f);
                RegisterFloat(PushSpeed);
                CreateSlider(PushSpeed, false);

                MovieFrameValue = new JSONStorableFloat("Movie Frame", 0f, OnSliderFrameChanged, -1.5f, 0.5f);
                RegisterFloat(MovieFrameValue);
                CreateSlider(MovieFrameValue, false);

                JSONStorableString postionTitleText = new JSONStorableString("Position Control",
                    "<color=#000><size=35><b>Position Control</b></size></color>");
                UIDynamic postionTitleTextfield = CreateTextField(postionTitleText, false);
                postionTitleTextfield.height = 30.0f;

                var person_btn = CreateButton("Person Update", false);

                person_btn.button.onClick.AddListener(() =>
                {
                    UpdateAtomData();
                });

                _personSelectJSON = new JSONStorableStringChooser("Target Choice", null, null, "", OnTargetPersonSelect);

                _personSelectJSON.storeType = JSONStorableParam.StoreType.Physical;
                RegisterStringChooser(_personSelectJSON);

                _personDisplayPopup = CreateScrollablePopup(_personSelectJSON, false);
                _personDisplayPopup.popupPanelHeight = 300f;
                _personDisplayPopup.popup.onOpenPopupHandlers += SyncPersonChoices;
                _personDisplayPopup.label = "Person:";

                _orificeJSON = new JSONStorableStringChooser("OrificeChoice", null, "Vagina", "", OrificeSelect);
                _orificeJSON.storeType = JSONStorableParam.StoreType.Physical;
                RegisterStringChooser(_orificeJSON);

                _displayPopup = CreateScrollablePopup(_orificeJSON, false);
                _displayPopup.popupPanelHeight = 300f;
                _displayPopup.popup.onOpenPopupHandlers += SyncOrificeChoices;
                _displayPopup.label = "Target Orifice:";

                _orificeJSON.choices = _orificePersonChoices;

                BodyRotationValue = new JSONStorableFloat("Body Rotation", 0f, OnBodyRotationChanged, -180f, 180f);
                RegisterFloat(BodyRotationValue);
                CreateSlider(BodyRotationValue, false);

                var btn = CreateButton("Target Tracking", false);
                btn.button.onClick.AddListener(() =>
                {
                    //OnTargetTracking();
                    OnBodyPosChanged(_bodyPosValue.val);
                });

                JSONStorableAction tracking_action = new JSONStorableAction("SetTracking", TriggerSyncReset);
                RegisterAction(tracking_action);

                _bodyPosValue = new JSONStorableFloat("Body Depth", 0f, OnBodyPosChanged, 0f, 1f);
                RegisterFloat(_bodyPosValue);
                CreateSlider(_bodyPosValue, false);

                _bodyYCorrection = new JSONStorableFloat("Body Y pos", 0f, OnBobyYCorrection, -0.5f, 0.5f);
                RegisterFloat(_bodyYCorrection);
                CreateSlider(_bodyYCorrection, false);

                JSONStorableString appearanceTitleText = new JSONStorableString("Appearance",
                    "<color=#000><size=35><b>Appearance</b></size></color>");
                UIDynamic appearanceTitleTextfield = CreateTextField(appearanceTitleText, true);
                appearanceTitleTextfield.height = 30.0f;

                GenitaliaPowerValue = new JSONStorableFloat("Genitalia Scale", 1f, OnSliderGenitaliaScaleChanged, 0.8f, 1.7f);
                RegisterFloat(GenitaliaPowerValue);
                CreateSlider(GenitaliaPowerValue, true);

                GenitaliaRocValue = new JSONStorableFloat("Genitalia Rotation", 0f, OnSliderGenitaliaRocChanged, -10f, 45f);
                RegisterFloat(GenitaliaRocValue);
                CreateSlider(GenitaliaRocValue, true);

                BodyAngleLock = new JSONStorableBool("Body Angle Lock", false, ChangeBodyAngleLock);
                RegisterBool(BodyAngleLock);
                UIDynamicToggle toggle4 = CreateToggle(BodyAngleLock, true);
                toggle4.label = "Body Angle Lock";

                BodyAlphaValue = new JSONStorableFloat("Body Alpha", 1f, OnSliderBodyAlphaChanged, 0f, 1f);
                RegisterFloat(BodyAlphaValue);
                CreateSlider(BodyAlphaValue, true);

                HSVColor ghsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                BodyColorValue = new JSONStorableColor("Body Color", ghsvc, OnBodyColorChanged);
                RegisterColor(BodyColorValue);
                CreateColorPicker(BodyColorValue, true);

                JSONStorableString targetTitleText = new JSONStorableString("Appearance",
                    "<color=#000><size=35><b>Head Target</b></size></color>");
                UIDynamic targetTitleTextfield = CreateTextField(targetTitleText, true);
                targetTitleTextfield.height = 30.0f;

                IsHeadActive = new JSONStorableBool("Active", false, SetTargetActive);
                RegisterBool(IsHeadActive);
                UIDynamicToggle toggle3 = CreateToggle(IsHeadActive, true);
                toggle3.label = "Head On Off";

                _headTargetBodyChoices = new List<string>();

                var btn2 = CreateButton("Head Target Update", true);

                btn2.button.onClick.AddListener(() =>
                {
                    OnTargetListUpdate();
                });

                _headTargetSelectJSON = new JSONStorableStringChooser("Head Choice", null, null, "", OnTargetSelect);
                _headTargetSelectJSON.storeType = JSONStorableParam.StoreType.Physical;
                RegisterStringChooser(_headTargetSelectJSON);

                _headTargetDisplayPopup = CreateScrollablePopup(_headTargetSelectJSON, true);
                _headTargetDisplayPopup.popupPanelHeight = 300f;
                _headTargetDisplayPopup.popup.onOpenPopupHandlers += SyncOrificeChoices;
                _headTargetDisplayPopup.label = "Target:";

                //_headTargetSelectJSON.choices = _headTargetBodyChoices;

                headXRotation = new JSONStorableFloat("head x rotation", 0f, OnChangeXRotation, -180f, 180f);
                RegisterFloat(headXRotation);
                CreateSlider(headXRotation, true);

                headYRotation = new JSONStorableFloat("head y rotation", 0f, OnChangeXRotation, -180f, 180f);
                RegisterFloat(headYRotation);
                CreateSlider(headYRotation, true);

                headZRotation = new JSONStorableFloat("head z rotation", 0f, OnChangeXRotation, -180f, 180f);
                RegisterFloat(headZRotation);
                CreateSlider(headZRotation, true);

                // A tiny bit of help to give some info to your fellow creator and/or player
                JSONStorableString helpText = new JSONStorableString("About",
                    "<color=#000><size=35><b>Glass Body</b></size></color>\n\n" +
                    "<color=#333>" +
                    "Move or vibrate the asset.\n\n" +
                    "thank you for your support\n" +
                    "<b>https://www.patreon.com/feel_vam</b>\n\n" +
                    "</color>");
                UIDynamic helpTextfield = CreateTextField(helpText, true);
                helpTextfield.height = 200.0f;

                var btn3 = CreateButton("Go To Support", true);
                btn3.button.onClick.AddListener(() =>
                {
                    Application.OpenURL("https://www.patreon.com/feel_vam");
                });


                UpdateAtomData();
                OnTargetListUpdate();

                SuperController.LogMessage(">> " + _personSelectJSON.val);
                SuperController.LogMessage(">> " + _headTargetSelectJSON.val);

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // Update is called with each rendered frame by Unity
        void Update()
        {
            // I don't know why the fuck this asset does not get the children properly during the loading process, so I'm checking the singleton
            // In reality, that's even cleaner to do it that way
            if (initLoops < maxInitLoops && editorInitialized == false && !SuperController.singleton.isLoading)
            {
                initLoops++;
                InitEditor();
            }

            if (initLoops == maxInitLoops)
            {
                SuperController.LogError("GlassBody : This CUA could not be initialized properly. Please select a compatible object and reload the script.");
                initLoops++; // To avoid spamming the log with the error
            }

            // first set.
            if(first_loading == false && BodyMeshRoot != null)
            {
                first_loading = true;

                OnSliderBodyAlphaChanged(BodyAlphaValue.val);
                OnBodyColorChanged(BodyColorValue);
                OnSliderGenitaliaScaleChanged(GenitaliaPowerValue.val);
                
            }
            
            
        }

        void FixedUpdate()
        {
            if (AbdomenRoot)
            {
                if (IsMovie.val)
                {
                    float frame_value = (PushMaxLevel.val + PushLimLevel.val) / (100f * PushSpeed.val);
                    if (is_movie_up)
                    {
                        if (PushLimLevel.val > curr_movie_value)
                            curr_movie_value += frame_value;
                        else
                            is_movie_up = false;
                    }
                    else
                    {
                        if ((PushMaxLevel.val * -1f) < curr_movie_value)
                            curr_movie_value -= frame_value;
                        else
                            is_movie_up = true;

                    }
                    //SuperController.LogMessage(">> " + curr_movie_value);
                    if (PushLimLevel.val == 0f && PushMaxLevel.val == 0f) curr_movie_value = 0f;

                    float ab_value = 0.0008f * curr_movie_value - 0f;
                    //ChestRoot.localPosition = new Vector3(-1.446327E-18f, 0.00181779f, 9.313226E-12f + ab_value);
                    float angle = 22f * (curr_movie_value / 2f);
                    SetPushAngle(angle);
                }
            }

            if (HeadTargetRoot)
            {
                if(IsHeadActive.val && _headTargetSelectJSON.val != "")
                {
                    //SuperController.LogMessage(">> " + _headTargetSelectJSON.val);
                    HeadTargetRoot.position = CumHeadRoot.position;

                    HeadTargetRoot.transform.rotation = CumHeadRoot.rotation;
                    HeadTargetRoot.transform.Rotate(headXRotation.val - 90f, headYRotation.val, headZRotation.val);
                }
            }
        }

        void OnDestroy()
        {
            
        }

        private bool InitEditor()
        {
            if (editorInitialized == true) return true;

            ObjectRoot = getObjectRoot(containingAtom.reParentObject, "GlassBody");
            if (ObjectRoot == null)
            {
                return false;
            }

            ArmatureRoot = getChildRoot(containingAtom.reParentObject, "Armature");

            LegRoot = getChildRoot(containingAtom.reParentObject, "Leg");
            //RightLegRoot = getChildRoot(containingAtom.reParentObject, "RightLeg");
            ChestRoot = getChildRoot(containingAtom.reParentObject, "Chest");
            AbdomenRoot = getChildRoot(containingAtom.reParentObject, "Abdomen");
            GenitaliaRoot = getChildRoot(containingAtom.reParentObject, "Genitalia");
            Gen1Root = getChildRoot(containingAtom.reParentObject, "Gen1");
            Gen2Root = getChildRoot(containingAtom.reParentObject, "Gen2");
            Gen3Root = getChildRoot(containingAtom.reParentObject, "Gen3");
            Gen4Root = getChildRoot(containingAtom.reParentObject, "Gen4");
            Gen5Root = getChildRoot(containingAtom.reParentObject, "Gen5");
            Gen6Root = getChildRoot(containingAtom.reParentObject, "Gen6");
            Gen7Root = getChildRoot(containingAtom.reParentObject, "Gen7");
            Gen8Root = getChildRoot(containingAtom.reParentObject, "Gen8");

            BodyMeshRoot = getChildRoot(containingAtom.reParentObject, "Body");
            CumHeadRoot = getChildRoot(containingAtom.reParentObject, "CumEffect");

            //SuperController.LogMessage(">> ----------------------------");

            // I found my transforms, now the rest
            if (ArmatureRoot && AbdomenRoot)
            {

                for (int i = 0; i < 8; i++)
                {
                    Transform GenRoot = getChildRoot(containingAtom.reParentObject, "Gen" + (i + 1));
                    base_pos[i] = GenRoot.localPosition;
                }

                editorInitialized = true;

                // Everything is initialized, triggering the callbacks to init the door rotation saved in the scene file
                // I'm just quickly fixing the script for VamTastic, I'm not gonna refactor everything, so in reality,
                // the float val param of the callbacks is useless

                return true;
            }
            else
            {
                SuperController.LogError("GlassBody : your CUA is missing mandatory components.");
                initLoops = maxInitLoops; // We don't have what we need, we consider the initialization failed
                return false;
            }

            return false;
        }

        // Recursive function searching for a transform containing the string set in objectNameSearchString
        // It is used the first time for the initialization of the script
        // This script NEEDS to use a string "containing" your original because the cloning of the original asset will add "Clone" behind it.
        private Transform getObjectRoot(Transform parent, string rootName)
        {
            foreach (Transform child in parent)
            {
                if (child.name.Contains(rootName))
                {
                    return child;
                }
                else
                {
                    Transform childSearch = getObjectRoot(child, rootName);
                    if (childSearch != null) return childSearch;
                }
            }

            return null;
        }

        private Transform getChildRoot(Transform parent, string searchName)
        {
            foreach (Transform child in parent)
            {
                if (child.name == searchName)
                {

                    return child;
                }
                else
                {
                    Transform childSearch = getChildRoot(child, searchName);
                    if (childSearch != null) return childSearch;
                }
            }

            return null;
        }

        // Right Door
        void OnSliderMaxChanged(float value)
        {
        }

        void OnSliderLimChanged(float value)
        {
        }

        void OnSliderSpeedChanged(float value)
        {
            
        }


        protected void OnTargetPersonSelect(string target_name)
        {
            SuperController.LogMessage(">> OnTargetPersonSelect " + target_name);
            if (target_name != null && target_name != "")
            {
                _selectAtom = SuperController.singleton.GetAtomByUid(target_name);
                _lipTrigger = _selectAtom.rigidbodies.First(rb => rb.name == "LipTrigger");
                _mouthTrigger = _selectAtom.rigidbodies.First(rb => rb.name == "MouthTrigger");
                _labiaTrigger = _selectAtom.rigidbodies.First(rb => rb.name == "LabiaTrigger");
                _vaginaTrigger = _selectAtom.rigidbodies.First(rb => rb.name == "VaginaTrigger");

                //foreach (Rigidbody rigidbody in _selectAtom.rigidbodies)
                //{
                //    SuperController.LogMessage(">> OnTargetPersonSelect " + rigidbody.name);
                //}

                //SuperController.LogMessage(">> OnTargetPersonSelect " + _vaginaTrigger);
            }

        }

        protected void SyncPersonChoices()
        {
            //SuperController.LogMessage(">> SyncPersonChoices ");
        }

        private void UpdateAtomData()
        {
            _atomBodyChoices = GetAtomListChoices();
            _personSelectJSON.choices = _atomBodyChoices;
        }

        protected List<string> GetAtomListChoices()
        {
            List<string> atomChoices = new List<string>();
            Atom atom;
            foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
            {
                atom = SuperController.singleton.GetAtomByUid(atomUID);
                JSONStorable geometry = atom.GetStorableByID("geometry");
                if (geometry != null)
                {
                    atomChoices.Add(atomUID);
                    //SuperController.LogMessage(">>" + atomUID);
                }
            }
            return atomChoices;
        }

        void OnChangeXRotation(float value)
        {

        }


        protected void SyncOrificeChoices()
        {
            
        }

        protected void OrificeSelect(string target_name)
        {
            //SuperController.LogMessage(">> " + target_name);
        }

        protected void OnBodyRotationChanged(float value)
        {

        }



        protected void OnTargetTracking()
        {
            //SuperController.LogMessage(">> OnTargetTracking " + _orificeJSON.val);


            bool isl = SuperController.singleton.isLoading;
            if (!isl)
            {
                if (_labiaTrigger != null)
                {
                    if (_orificeJSON.val == "Mouth")
                    {
                        _thisControl.transform.position = _lipTrigger.transform.position + new Vector3(0f, _bodyYCorrection.val, 0f);
                        _thisControl.transform.rotation = Quaternion.LookRotation(_mouthTrigger.transform.position - _lipTrigger.transform.position, _lipTrigger.transform.up);

                    }
                    if (_orificeJSON.val == "Vagina")
                    {
                        _thisControl.transform.position = _labiaTrigger.transform.position + new Vector3(0f, _bodyYCorrection.val, 0f);
                        _thisControl.transform.rotation = Quaternion.LookRotation(_vaginaTrigger.transform.position - _labiaTrigger.transform.position, _labiaTrigger.transform.up);
                    }
                    if (_orificeJSON.val == "Anus")
                    {
                        Transform anusTransform = _thisControl.transform;

                        Vector3 anusOffsetPosition = new Vector3(0.0f, -0.025f, 0.008f);
                        
                        _thisControl.transform.position = _labiaTrigger.transform.position + new Vector3(0f, _bodyYCorrection.val, 0f);
                        _thisControl.transform.rotation = Quaternion.LookRotation(_vaginaTrigger.transform.position - anusTransform.position, anusTransform.up);

                        anusTransform.Translate(anusOffsetPosition, anusTransform);
                    }

                    Vector3 offsetPosition = new Vector3(0f, 0f, -0.3f);
                    _thisControl.transform.Translate(offsetPosition, _thisControl.transform);

                    _thisControl.transform.Rotate(new Vector3(0f, 0f, BodyRotationValue.val));

                }
            }
        }


        protected void SetTargetActive(JSONStorableBool jb)
        {
            if (!jb.val)
            {


            }
        }

        protected void ChangePushMovie(JSONStorableBool jb)
        {
            if(!jb.val)
            {
                SetPushAngle(0f);

            }
        }

        void OnSliderFrameChanged(float value)
        {
            float angle = 22f * (value / 2f);
            SetPushAngle(angle);
        }

        void OnBodyPosChanged(float value)
        {
            if(_labiaTrigger != null)
            {
                OnTargetTracking();
                Vector3 offsetPosition = new Vector3(0f, 0f, 0.3f * value);
                _thisControl.transform.Translate(offsetPosition, _thisControl.transform);
            }
            
        }

        void OnBobyYCorrection(float value)
        {
            OnBodyPosChanged(_bodyPosValue.val);
        }

        void OnSliderGenitaliaStateChanged(float value)
        {
            if (GenitaliaRoot == null) return;
            for (int i = 0; i < 8; i++)
            {
                Transform GenRoot = getChildRoot(containingAtom.reParentObject, "Gen" + (i + 1));
                FixedJoint fixedjoint = GenRoot.GetComponent<FixedJoint>();
                fixedjoint.massScale = value;
            }

            SuperController.LogMessage(">> OnSliderGenitaliaStateChanged " + value);
        }

        void OnSliderBodyAlphaChanged(float value)
        {
            if(BodyMeshRoot != null)
            {
                Renderer renderer = BodyMeshRoot.GetComponent<Renderer>();
                Color cc = renderer.material.GetColor("_Color");
                cc.a = value;
                renderer.material.SetColor("_Color", cc);

                if (value < 1f)
                    ChangeRenderMode(BodyMeshRoot, "Transparent");
                else
                    ChangeRenderMode(BodyMeshRoot, "Opaque");
            }
        }

        void OnBodyColorChanged(JSONStorableColor dc)
        {
            if(BodyMeshRoot != null)
            {
                Color c = Color.HSVToRGB(dc.valNoCallback.H, dc.valNoCallback.S, dc.valNoCallback.V);
                Renderer renderer = BodyMeshRoot.GetComponent<Renderer>();
                c.a = renderer.material.color.a;
                renderer.material.SetColor("_Color", c);
            }
        }

        void OnSliderGenitaliaRocChanged(float value)
        {
            if (GenitaliaRoot == null) return;

            GenitaliaRoot.localRotation = Quaternion.Euler(0f, 0f, -90f - value);
            if(BodyAngleLock.val) ArmatureRoot.localRotation = Quaternion.Euler(-90f + value, 0f, 0f);
        }

        protected void ChangeBodyAngleLock(JSONStorableBool jb)
        {
            if (!jb.val)
            {
                ArmatureRoot.localRotation = Quaternion.Euler(-90f, 0f, 0f);
            }
        }


        void OnSliderGenitaliaScaleChanged(float value)
        {
            if (GenitaliaRoot == null) return;

            for (int i = 0; i < 8; i++)
            {
                Transform GenRoot = getChildRoot(containingAtom.reParentObject, "Gen" + (i + 1));
                Rigidbody rigidbody = GenRoot.GetComponent<Rigidbody>();
                //rigidbody.useGravity = false;
                //rigidbody.isKinematic = true;
                if(GenRoot.GetComponent<SphereCollider>() != null) GenRoot.GetComponent<SphereCollider>().enabled = false;
                FixedJoint fixedjoint = GenRoot.GetComponent<FixedJoint>();
                Destroy(fixedjoint);

                GenRoot.localPosition = base_pos[i];
            }

            Gen1Root.localScale = new Vector3(value, value, value);

            Rigidbody PrevRoot = GenitaliaRoot.GetComponent<Rigidbody>();
            for (int i = 0; i < 8; i++)
            {
                Transform GenRoot = getChildRoot(containingAtom.reParentObject, "Gen" + (i + 1));
                FixedJoint fixedjoint = GenRoot.gameObject.AddComponent<FixedJoint>();
                fixedjoint.connectedBody = PrevRoot;
                fixedjoint.massScale = 1.5f;

                //GenRoot.GetComponent<SphereCollider>().enabled = true;
                //Rigidbody rigidbody = GenRoot.GetComponent<Rigidbody>();
                //rigidbody.drag = 60f;
                //rigidbody.useGravity = true;

                PrevRoot = GenRoot.GetComponent<Rigidbody>();
            }

            if (_inter_value != null) StopCoroutine(_inter_value);
            _inter_value = InterValue();
            StartCoroutine(_inter_value);


        }

        private IEnumerator _inter_value;

        IEnumerator InterValue()
        {
            //yield return new WaitForEndOfFrame();
            yield return new WaitForSeconds(0.5f);

            for (int i = 0; i < 8; i++)
            {
                Transform GenRoot = getChildRoot(containingAtom.reParentObject, "Gen" + (i + 1));
                if(GenRoot.GetComponent<SphereCollider>() != null) GenRoot.GetComponent<SphereCollider>().enabled = true;
                Rigidbody rigidbody = GenRoot.GetComponent<Rigidbody>();
                rigidbody.drag = 60f;
                rigidbody.mass = 5f;
                //rigidbody.isKinematic = false;
                //rigidbody.useGravity = true;

            }

            //SuperController.LogMessage(">> InterValue ");
        }

        protected void OnTargetSelect(string target_name)
        {
            SuperController.LogMessage(">> OnTargetSelect " + target_name);
            if (_headTargetBodyChoices.Count > 0) OnSelectTargetObj(target_name);
        }

        private void OnSelectTargetObj(string uid)
        {
            Atom atom = SuperController.singleton.GetAtomByUid(uid);
            FreeControllerV3 thisControl = atom.freeControllers.First(fc => fc.name == "control");
            HeadTargetRoot = thisControl.transform;

        }

        private void OnTargetListUpdate()
        {
            List<string> uids = SuperController.singleton.GetAtomUIDs();

            _headTargetBodyChoices = new List<string>();

            foreach (string uid in uids)
            {
                Atom atom = SuperController.singleton.GetAtomByUid(uid);

                if(atom.type == "CustomUnityAsset" && _thisAtom.name != atom.name)
                {
                    _headTargetBodyChoices.Add(uid);
                }
            }

            _headTargetSelectJSON.choices = _headTargetBodyChoices;
        }


        void SetPushAngle(float value)
        {
            if (AbdomenRoot == null) return;
            if (LegRoot == null) return;
            if (GenitaliaRoot == null) return;

            AbdomenRoot.localRotation = Quaternion.Euler(0f, 90f, -88f + value);
            LegRoot.localRotation = Quaternion.Euler(0f, 0f, 6f - (value * 1.5f));
            GenitaliaRoot.localRotation = Quaternion.Euler(0f, 0f, -90f - (value + GenitaliaRocValue.val));
            //ArmatureRoot.localRotation = Quaternion.Euler(-90f + value, 0f, 0f);
            ChestRoot.localRotation = Quaternion.Euler(90f + (value * 0.1f), 0f, 0f);
        }

        protected void ChangeRenderMode(Transform target, string blendMode)
        {
            if (target == null)
            {
                return;
            }
            Renderer renderer = target.GetComponent<Renderer>();

            if (blendMode == "Opaque")
            {
                renderer.material.SetFloat("_Mode", 0.0f);
                renderer.material.SetOverrideTag("RenderType", "Opaque");
                renderer.material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
                renderer.material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.Zero);
                renderer.material.SetInt("_ZWrite", 1);
                renderer.material.DisableKeyword("_ALPHATEST_ON");
                renderer.material.DisableKeyword("_ALPHABLEND_ON");
                renderer.material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
                renderer.material.renderQueue = -1;
            }
            else
            {
                renderer.material.SetFloat("_Mode", 3.0f);
                renderer.material.SetOverrideTag("RenderType", "Transparent");
                renderer.material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.One);
                renderer.material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
                renderer.material.SetInt("_ZWrite", 0);
                renderer.material.DisableKeyword("_ALPHATEST_ON");
                renderer.material.DisableKeyword("_ALPHABLEND_ON");
                renderer.material.EnableKeyword("_ALPHAPREMULTIPLY_ON");
                renderer.material.renderQueue = 3000;
            }
        }

        private void TriggerSyncReset()
        {
            OnBodyPosChanged(_bodyPosValue.val);
        }
    }

}